"""Contains strategies and order execution functionality."""

# For deep-copied dictionaries
from copy import deepcopy

# For managing dates
from datetime import datetime, timedelta

# For managing tables properly
from numpy import where
from pandas import DataFrame, read_csv, read_excel, date_range, Timedelta, Timestamp

# For mathematical operations
from math import isnan

# For nice-looking progress bars
from tqdm import tqdm

# For white noise strategy
from random import choice

from _helper_functions import find_dataframe_value_with_keywords, \
    find_price, calculate_portfolio_value, calculate_relative_gross_exposure
from _1_data_preparation import save_dataframe_to_csv
from _3_performance_evaluation import evaluate_performance


def initialize_trading_journal():
    """Initializes a pandas DataFrame that serves as a trading journal.

    Initialization is important for determining the column order.
    """
    df_trading_journal = DataFrame(columns=[
        'datetime',
        'Cash',
        'Cash before',
        'Asset',
        'Buy or sell',
        'Number bought',
        'Price (quote without any fees)',
        'Value bought',
        'Portfolio value',
        'Dict of assets in portfolio',
        'Absolute fees (as absolute)',
        'Current equity margin',
        'Exposure (in currency)',
        'Exposure (number)',
        'Gross exposure',
        'Interest paid',
        'Money spent',
        'Relative fees (as absolute)',
        'Relative fees (as relative)',
        'Strategy ID',
        'Total exposure',
        'Total fees (as absolute)',
        'Total fees (as relative)'
    ])

    return df_trading_journal

def execute_order(
        boolean_buy,
        index,
        date,
        crypto_key,
        number_to_be_bought,
        strategy_id,
        df_prices,
        df_trading_journal,
        margin_loan_rate,
        fees,
        float_budget_in_usd,
        price,
        display_options,
        constraints,
        general_settings,
        boolean_allow_partially_filled_orders
    ):
    """Executes all kinds of orders.

    Can handle more than one order per point in time by subsequently calling
    this function.
    """
    if boolean_buy and (number_to_be_bought < 0) or not boolean_buy and (number_to_be_bought > 0):
        raise ValueError(f'boolean_buy: {boolean_buy} and number_to_be_bought: {number_to_be_bought} is contradictory.')

    order = {
        'datetime': date,
        'Strategy ID': strategy_id,
        'Asset': crypto_key,
        'Buy or sell': boolean_buy
    }

    if len(df_trading_journal) > 0:
        available_funds = df_trading_journal['Cash'].iloc[-1] - constraints['minimum_cash']
        order['Dict of assets in portfolio'] = deepcopy(df_trading_journal['Dict of assets in portfolio'].iloc[-1])
        order['Cash before'] = df_trading_journal['Cash'].iloc[-1]
    else:
        available_funds = float_budget_in_usd
        order['Dict of assets in portfolio'] = {crypto_key: 0}
        order['Cash before'] = float_budget_in_usd

    def reduce_quantity_until_max_gross_exposure_is_met(
            number_to_be_bought,
            df_prices,
            dict_of_assets_in_portfolio,
            time,
            display_options,
            constraints,
            rounding_accuracy,
            cash_value,
            general_settings,
            crypto_key
        ):
        # Checks if the maximum_gross_exposure constraint is violated and
        # reduces the order volumne step-by-step in case of a violation
        # until the exposure falls within the constraint.
        initial_number_to_be_bought = number_to_be_bought
        dict_of_assets_in_portfolio[crypto_key] = dict_of_assets_in_portfolio[crypto_key] + number_to_be_bought
        for reduction_step in range(1, 100):
            dict_of_assets_in_portfolio = deepcopy(dict_of_assets_in_portfolio)
            relative_gross_exposure = calculate_relative_gross_exposure(
                df_prices=df_prices,
                dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
                time=time,
                display_options=display_options,
                constraints=constraints,
                rounding_accuracy=rounding_accuracy,
                cash_value=cash_value,
                general_settings=general_settings
            )

            # Stop reduction as soon as constraint is met, reduce if
            # constraint is violated.
            if relative_gross_exposure <= constraints['maximum_gross_exposure']:
                break
            else:
                dict_of_assets_in_portfolio[crypto_key] = ((100 - reduction_step) / 100) * initial_number_to_be_bought

        return initial_number_to_be_bought

    def quantity_that_can_be_bought_given_budget(order, dict_of_assets_in_portfolio, boolean_allow_partially_filled_orders, number_to_be_bought, available_funds, general_settings, display_options, date):
        initial_number_to_be_bought = number_to_be_bought

        number_to_be_bought = min(
            max(round(available_funds / price, general_settings['rounding_decimal_places_for_security_quantities']), 0),
            number_to_be_bought
        )

        # Todo: Individual asset constraints.
        # constraints['maximum_individual_asset_exposure_all']
        dict_of_assets_in_portfolio = deepcopy(dict_of_assets_in_portfolio)
        number_to_be_bought = reduce_quantity_until_max_gross_exposure_is_met(
            number_to_be_bought=number_to_be_bought,
            df_prices=df_prices,
            dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
            time=date,
            display_options=display_options,
            constraints=constraints,
            rounding_accuracy=general_settings['rounding_decimal_places'],
            cash_value=order['Cash before'] - (number_to_be_bought * price) - fees['absolute_fee_buy_order'] - round(
                fees['percentage_buying_fees_and_spread'] * price * number_to_be_bought,
                general_settings['rounding_decimal_places']
            ),
            general_settings=general_settings,
            crypto_key=crypto_key
        )

        if boolean_allow_partially_filled_orders:
            if number_to_be_bought == 0 and initial_number_to_be_bought != 0:
                if display_options['warning_buy_order_could_not_be_filled']:
                    print(f'Order execution warning: Buy order for {initial_number_to_be_bought} units of {crypto_key} could not be filled.')
            elif number_to_be_bought < initial_number_to_be_bought:
                if display_options['warning_buy_order_could_not_be_filled']:
                    print(f'Order execution warning: Buy order for {initial_number_to_be_bought} units of {crypto_key} could only partially be filled: {number_to_be_bought} units bought.')
            elif number_to_be_bought > initial_number_to_be_bought:
                raise ValueError(f'It is not possible that the signal quantity {initial_number_to_be_bought} is lower than the final quantity {number_to_be_bought} for {order["Asset"]}.')
            return number_to_be_bought
        else:
            if initial_number_to_be_bought != number_to_be_bought:
                raise InputError('Quantity {number_to_be_bought} for {crypto_key} cannot be covered with the given budget or constraints. Maximum {max_possible_quantity} units can be bought.')
            else:
                return number_to_be_bought

    def quantity_that_can_be_sold_given_portfolio(order, dict_of_assets_in_portfolio, boolean_allow_partially_filled_orders, number_to_be_bought, df_trading_journal, general_settings, display_options, date):
        initial_number_to_be_bought = number_to_be_bought

        # Todo: Individual asset constraints.
        # constraints['maximum_individual_asset_exposure_all']

        positive_quantity = (-1) * number_to_be_bought
        try:
            number_to_be_bought = (-1) * min(
                positive_quantity,
                df_trading_journal['Dict of assets in portfolio'].iloc[-1][crypto_key]
            )
        except IndexError:
            number_to_be_bought = 0

        dict_of_assets_in_portfolio = deepcopy(order['Dict of assets in portfolio'])
        number_to_be_bought = reduce_quantity_until_max_gross_exposure_is_met(
            number_to_be_bought=number_to_be_bought,
            df_prices=df_prices,
            dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
            time=date,
            display_options=display_options,
            constraints=constraints,
            rounding_accuracy=general_settings['rounding_decimal_places'],
            cash_value=order['Cash before'] - (number_to_be_bought * price) - fees['absolute_fee_buy_order'] - round(
                fees['percentage_buying_fees_and_spread'] * price * number_to_be_bought,
                general_settings['rounding_decimal_places']
            ),
            general_settings=general_settings,
            crypto_key=crypto_key
        )

        if boolean_allow_partially_filled_orders:
            if number_to_be_bought == 0 and initial_number_to_be_bought != 0:
                if display_options['warning_buy_order_could_not_be_filled']:
                    print(f'Order execution warning: Buy order for {number_to_be_bought} units of {crypto_key} could not be filled.')
            elif number_to_be_bought > initial_number_to_be_bought:
                if display_options['warning_buy_order_could_not_be_filled']:
                    print(f'Order execution warning: Buy order for {number_to_be_bought} units of {crypto_key} could not be filled.')
            elif number_to_be_bought < initial_number_to_be_bought:
                raise ValueError('It is not possible that the signal quantity {initial_number_to_be_bought} is higher (less assets sold) than the final quantity {number_to_be_bought} for {order["Asset"]}.')
            return number_to_be_bought
        else:
            if number_to_be_bought != initial_number_to_be_bought:
                raise InputError('Quantity {number_to_be_bought} for {crypto_key} cannot be sold with the given assets or constraints. Maximum {number_to_be_bought} units can be sold.')
            else:
                return number_to_be_bought

    if boolean_buy:
        number_to_be_bought = quantity_that_can_be_bought_given_budget(
            order=order,
            dict_of_assets_in_portfolio=order['Dict of assets in portfolio'],
            boolean_allow_partially_filled_orders=boolean_allow_partially_filled_orders,
            number_to_be_bought=number_to_be_bought,
            available_funds=available_funds,
            general_settings=general_settings,
            display_options=display_options,
            date=date
        )
    else:
        number_to_be_bought = quantity_that_can_be_sold_given_portfolio(
            order=order,
            dict_of_assets_in_portfolio=order['Dict of assets in portfolio'],
            boolean_allow_partially_filled_orders=boolean_allow_partially_filled_orders,
            number_to_be_bought=number_to_be_bought,
            df_trading_journal=df_trading_journal,
            general_settings=general_settings,
            display_options=display_options,
            date=date
        )

    if isnan(price):
        number_to_be_bought = 0
        price = 0

    order['Dict of assets in portfolio'][crypto_key] = order['Dict of assets in portfolio'][crypto_key] + number_to_be_bought

    if number_to_be_bought != 0:
        if boolean_buy:
            order['Absolute fees (as absolute)'] = fees['absolute_fee_buy_order']
            order['Relative fees (as absolute)'] = round(
                fees['percentage_buying_fees_and_spread'] * price * number_to_be_bought,
                general_settings['rounding_decimal_places']
            )
            order['Relative fees (as relative)'] = fees['percentage_buying_fees_and_spread']
            order['Total fees (as absolute)'] = fees['absolute_fee_buy_order'] + order['Relative fees (as absolute)']
            order['Total fees (as relative)'] = round(
                order['Total fees (as absolute)'] / (number_to_be_bought * price),
                general_settings['rounding_decimal_places']
            )
        else:
            order['Absolute fees (as absolute)'] = fees['absolute_fee_sell_order']
            order['Relative fees (as absolute)'] = round(
                fees['percentage_selling_fees_and_spread'] * price * number_to_be_bought * (-1),
                general_settings['rounding_decimal_places']
            )
            order['Relative fees (as relative)'] = fees['percentage_selling_fees_and_spread']
            order['Total fees (as absolute)'] = fees['absolute_fee_sell_order'] + order['Relative fees (as absolute)']
            order['Total fees (as relative)'] = round(
                order['Total fees (as absolute)'] / (number_to_be_bought * price) * (-1),
                general_settings['rounding_decimal_places']
            )
    else:
        order['Absolute fees (as absolute)'] = 0
        order['Relative fees (as absolute)'] = 0
        order['Relative fees (as relative)'] = 0
        order['Total fees (as absolute)'] = 0
        order['Total fees (as relative)'] = 0

    if len(df_trading_journal) > 0:
        # Date conversion from string to date format; datetime is in microsecond format
        previous_time = df_trading_journal['datetime'].iloc[-1]
        days_since_last_order = date - previous_time

        order['Number bought'] = number_to_be_bought
        order['Value bought'] = round(price * number_to_be_bought, general_settings['rounding_decimal_places'])
        if days_since_last_order == timedelta(seconds=0):
            order['Interest paid'] = 0
        else:
            order['Interest paid'] = round(
                (
                    (1 - df_trading_journal['Current equity margin'].iloc[-1])
                    * (
                        df_trading_journal['Portfolio value'].iloc[-1]
                        * margin_loan_rate
                    )
                ) ** (
                    (
                        days_since_last_order.total_seconds() / 86400
                    ) / 365
                ),
                general_settings['rounding_decimal_places']
            ) # "/ 86400" because one day has 86400 seconds

        order['Money spent'] = round(
            number_to_be_bought * (
                price
            ) + (
                order['Total fees (as absolute)'] + order['Interest paid']
            ), general_settings['rounding_decimal_places']
        ) # Todo But everything that can be bought minus fees and other costs

        order['Cash'] = round(
            df_trading_journal['Cash'].iloc[-1] - order['Money spent'],
            general_settings['rounding_decimal_places']
        )

    else:
        # For initial row
        order['Number bought'] = number_to_be_bought
        order['Value bought'] = price * number_to_be_bought
        order['Interest paid'] = 0.0
        order['Money spent'] = number_to_be_bought * (
            price
        ) + (
            + order['Total fees (as absolute)']
            + order['Interest paid']
        ) # Todo But everything that can be bought minus fees and other costs

        order['Cash'] = round(
            float_budget_in_usd - order['Money spent'],
            general_settings['rounding_decimal_places']
        )

    order['Price (quote without any fees)'] = price

    # Todo: For now just 1
    order['Current equity margin'] = 1

    order['Exposure (number)'] = order['Dict of assets in portfolio'][order['Asset']]

    order['Exposure (in currency)'] = round(
        price * order['Exposure (number)'],
        general_settings['rounding_decimal_places']
    )

    order['Portfolio value'] = calculate_portfolio_value(
        df_prices=df_prices,
        dict_of_assets_in_portfolio=order['Dict of assets in portfolio'],
        time=order['datetime'],
        display_options=display_options,
        constraints=constraints,
        cash_value=order['Cash'],
        rounding_accuracy=general_settings['rounding_decimal_places']
    )

    order['Total exposure'] = round(
        order['Portfolio value'] - order['Cash'],
        general_settings['rounding_decimal_places']
    )

    assert round(order['Total exposure'], 2) == round(calculate_portfolio_value(
        df_prices=df_prices,
        dict_of_assets_in_portfolio=order['Dict of assets in portfolio'],
        time=order['datetime'],
        display_options=display_options,
        constraints=constraints,
        rounding_accuracy=general_settings['rounding_decimal_places']
    ), 2)

    order['Gross exposure'] = calculate_relative_gross_exposure(
        df_prices=df_prices,
        dict_of_assets_in_portfolio=order['Dict of assets in portfolio'],
        time=order['datetime'],
        display_options=display_options,
        constraints=constraints,
        rounding_accuracy=general_settings['rounding_decimal_places'],
        cash_value=order['Cash'],
        general_settings=general_settings
    )

    return order

def prepare_signal_list_ii(
        file_path_with_signal_data,
        strategy_hyperparameters
    ):
    try:
        try:
            df_signals = read_csv(
                file_path_with_signal_data,
                sep=',',
                parse_dates=True,
                infer_datetime_format=True,
                index_col=['datetime', 'id']
            )
        except UnicodeDecodeError:
            df_signals = read_excel(
                file_path_with_signal_data,
                sep=',',
                parse_dates=True,
                infer_datetime_format=True,
                index_col=['datetime', 'id']
            )
    except:
        try:
            df_signals = read_csv(
                'backtesting/' + file_path_with_signal_data,
                sep=',',
                parse_dates=True,
                infer_datetime_format=True,
                index_col=['datetime', 'id']
            )
        except:
            raise ValueError("Please setup a signal table. A signal table needs the following columns: 'datetime', 'id' (hexadecimal ERC20 token identifier), 'signal_strength' (numeric that is used to infer buy or sell orders)")

    # Only use top 50 tokens
    print(f'\nNumber of signals before dropping non-top-50 tokens: {len(df_signals)}')
    comments['Number of signals before dropping non-top-50 tokens'] = len(df_signals)
    print(f'Number of unique eth IDs before: {len(df_signals["id"].unique())}')
    comments['Number of unique eth IDs before'] = len(df_signals['id'].unique())
    df_top50 = read_csv('strategy_tables/Top50Tokens.csv', sep=',')
    list_top50 = df_top50['a.ID'].values.tolist()
    df_signals = df_signals[df_signals['id'].isin(list_top50)]
    print(f'Number of signals after dropping non-top-50 tokens: {len(df_signals)}')
    comments['Number of signals after dropping non-top-50 tokens'] = len(df_signals)
    print(f'Number of unique eth IDs after: {len(df_signals["id"].unique())}')
    comments['Number of unique eth IDs after'] = len(df_signals['id'].unique())

    # Drop weak signals
    print(f'\nNumber of signals before dropping weak signals: {len(df_signals)}')
    comments['Number of signals before dropping weak signals'] = len(df_signals)
    df_signals['signal_type'] = where(df_signals["rawSignal"]<=strategy_hyperparameters['sell_parameter'], "SELL", where(df_signals["rawSignal"]>=strategy_hyperparameters['buy_parameter'], "BUY", "HODL"))
    #indexNames = df_signals[(df_signals['rawSignal'] >= strategy_hyperparameters['sell_parameter']) & (df_signals['rawSignal'] <= strategy_hyperparameters['buy_parameter']) ].index # FOr testing purposes: 0.03 and 12.0
    #df_signals.drop(indexNames, inplace=True)
    df_signals = df_signals[~df_signals['signal_type'].isin(["HODL"])]
    print(f'Number of signals after dropping weak signals: {len(df_signals)}')
    comments['Number of signals after dropping weak signals'] = len(df_signals)

    # Drop signals that are not covered by ITSA
    print(f'\nNumber of signals before dropping unavailable data: {len(df_signals)}')
    comments['Number of signals before dropping unavailable data'] = len(df_signals)
    df_signals = df_signals[
        df_signals['id'].isin(df_prices['token_address_eth'].unique())
    ]
    print(f'Number of signals after dropping unavailable data: {len(df_signals)}')
    comments['Number of signals after dropping unavailable data'] = len(df_signals)

    # Drop data that is not needed
    print(f'\nNumber of data points before dropping unnecessary data: {len(df_prices)}')
    comments['Number of data points before dropping unnecessary data'] = len(df_prices)
    df_prices = df_prices[
        df_prices['token_address_eth'].isin(df_signals['id'].unique())
    ]
    print(f'Number of data points after dropping unnecessary data: {len(df_prices)}')
    comments['Number of data points after dropping unnecessary data'] = len(df_prices)

    # Drop signals that have assets that have no prices for the last day. This
    # is necessary because for the final portfolio valuation, there has to be a
    # price for the last day.
    print(f'\nNumber of signals before dropping unavailable prices: {len(df_signals)}')
    comments['Number of signals before dropping unavailable prices'] = len(df_signals)
    print(f'Number of unique eth IDs before: {len(df_signals["id"].unique())}')
    comments['Number of unique eth IDs before'] = len(df_signals['id'].unique())

    last_signal_date = df_signals['datetime'].iloc[-1]
    price = None

    list_of_assets_dropped_due_to_price_lag_at_last_day = []

    for asset_in_signal_table in df_prices['token_address_eth'].unique():
        price = find_price(
            df_prices,
            desired_index=index,
            boolean_allow_older_prices=False,
            boolean_allow_newer_prices=False,
            boolean_warnings=display_options['warning_no_price_for_last_day'],
            boolean_errors=False
        )

        if price is None:
            df_signals = df_signals[df_signals.ID != asset_in_signal_table]
            if display_options['warning_no_price_for_last_day']:
                print(f'{asset_in_signal_table} was dropped because there is no price for the last day. {price}')
            list_of_assets_dropped_due_to_price_lag_at_last_day.append(
                asset_in_signal_table
            )
        elif isnan(price):
            df_signals = df_signals[df_signals.ID != asset_in_signal_table]
            if display_options['warning_no_price_for_last_day']:
                print(f'{asset_in_signal_table} was dropped because there is only a NaN price for the last day. {price}')
            list_of_assets_dropped_due_to_price_lag_at_last_day.append(
                asset_in_signal_table
            )
        elif price == 0:
            df_signals = df_signals[df_signals.ID != asset_in_signal_table]
            if display_options['warning_no_price_for_last_day']:
                print(f'asset_in_signal_table was dropped because there is only a 0 price for the last day. {price}')
            list_of_assets_dropped_due_to_price_lag_at_last_day.append(
                asset_in_signal_table
            )

    df_signals = df_signals[~df_signals['id'].isin(list_of_assets_dropped_due_to_price_lag_at_last_day)]

    print(f'Number of assets that do not have a price on the last day: {len(list_of_assets_dropped_due_to_price_lag_at_last_day)}')
    comments['Number of assets that do not have a price on the last day'] = len(
        list_of_assets_dropped_due_to_price_lag_at_last_day
    )

    print(f'Number of signals after dropping unavailable prices: {len(df_signals)}')
    comments['Number of signals after dropping unavailable prices'] = len(df_signals)

    print(f'Number of unique eth IDs after: {len(df_signals["id"].unique())}')
    comments['Number of unique eth IDs after'] = len(df_signals['id'].unique())

    id_column_name = 'token_address_eth'

    for asset_with_possible_later_price in list_of_assets_dropped_due_to_price_lag_at_last_day:
        price = find_price(
            df_prices,
            asset_with_possible_later_price,
            time=datetime(2019, 7, 1),
            boolean_allow_older_prices=False,
            boolean_allow_newer_prices=True,
            boolean_warnings=display_options['warning_no_price_for_last_day']
        )
        if price is None:
            list_of_assets_dropped_due_to_price_lag_at_last_day.remove(
                asset_with_possible_later_price
            )

    print(f'Number of assets that do not have a price on the last day, but on a later day: {len(list_of_assets_dropped_due_to_price_lag_at_last_day)}')
    comments['Number of assets that do not have a price on the last day, but on a later day'] = len(list_of_assets_dropped_due_to_price_lag_at_last_day)

    return df_signals, id_column_name

def prepare_signal_list_san(
        file_path_with_signal_data,
        strategy_hyperparameters
    ):
    try:
        df_signals = read_csv(
            file_path_with_signal_data,
            sep=',',
            parse_dates=True,
            infer_datetime_format=True,
            index_col=['datetime', 'id']
        )
    except:
        raise ValueError("Please set up a signal table. A signal table needs the following columns: 'date' (yyyy-mm-dd), 'signal_strength' (numeric that is used to infer buy or sell orders), 'id' (hexadecimal ERC20 token identifier or ITIN). You can change the column names in the Excel/CSV file to fit this convention or in main.py so that the program adjusts to the naming in the table.")

    df_signals['signal_type'] = where(df_signals['signal_strength']<=strategy_hyperparameters['sell_parameter'],'SELL',where(df_signals['signal_strength']>=strategy_hyperparameters['buy_parameter'],'BUY','HODL'))
    df_signals = df_signals[~df_signals['signal_type'].isin(['HODL'])]

    return df_signals

def execute_strategy_multi_asset(
        df_prices,
        int_chosen_strategy,
        float_budget_in_usd,
        margin_loan_rate,
        boolean_allow_shorting,
        list_trading_execution_delay_after_signal_in_hours,
        dict_crypto_options,
        minimum_expected_mispricing_trigger_in_percent,
        strategy_hyperparameters,
        sell_at_the_end,
        file_path_with_signal_data,
        display_options,
        constraints,
        general_settings,
        comments
    ):
    """Filters signals and executes all remaining signals."""

    df_signals = prepare_signal_list_san(
        file_path_with_signal_data,
        strategy_hyperparameters
    )

    df_signals = df_signals.loc[strategy_hyperparameters['start_time']:strategy_hyperparameters['end_time'], ]

    if display_options['boolean_test']:
        # Drop non-test signals
        print('Warning: Test run!')
        print(f'\nNumber of signals before dropping non-test signals: {len(df_signals)}')
        comments['Number of signals before dropping non-test signals'] = len(df_signals)
        list_indexes_to_be_dropped = []
        df_signals = df_signals.sample(frac=0.05, random_state=0)
        df_signals.sort_index(inplace=True)
        print(f'Number of signals after dropping non-test signals: {len(df_signals)}')
        comments['Number of signals after dropping non-test signals'] = len(df_signals)

    df_trading_journal = initialize_trading_journal()

    for index, row in tqdm(df_signals.iterrows(), desc='Going through signals', unit='signal'):
        boolean_buy = None

        crypto_key = index[1]

        price = find_price(
            df_prices,
            desired_index=index,
            boolean_allow_older_prices=False,
            boolean_allow_newer_prices=False,
            boolean_warnings=display_options['warning_no_price_during_execution']
        )

        if price is not None and price > 0:
            if row['signal_type'] == "BUY":
                boolean_buy = True

                # Check if minimum is already crossed
                if len(df_trading_journal) > 0:
                    if df_trading_journal['Cash'].iloc[-1] <= constraints['minimum_cash']:
                        allow_buy_orders = False

                    else:
                        allow_buy_orders = True

                else:
                    allow_buy_orders = True

                if allow_buy_orders:
                    try:
                        float_budget_in_usd = df_trading_journal['Cash'].iloc[-1]
                    except:
                        pass

                    number_to_be_bought = round(
                        (
                            float_budget_in_usd
                            - constraints['minimum_cash']
                        ) / price,
                        general_settings['rounding_decimal_places_for_security_quantities']
                    )

                else:
                    number_to_be_bought = None

                if number_to_be_bought is not None:
                    number_to_be_bought = number_to_be_bought * strategy_hyperparameters['maximum_relative_exposure_per_buy']

            elif row['signal_type'] == "SELL":
                boolean_buy = False
                # Check if exposure for this asset is zero and skip order if so

                if not (
                    (
                        boolean_buy == False
                    ) and (
                        find_dataframe_value_with_keywords(
                            df_trading_journal,
                            search_term_1=crypto_key,
                            search_column_name_1='Asset'
                        )
                    ) is None
                ):
                    number_to_be_bought = (-1) * (
                        find_dataframe_value_with_keywords(
                            df_trading_journal,
                            search_term_1=crypto_key,
                            search_column_name_1='Asset',
                            output_column_name='Exposure (number)',
                            first_last_or_all_elements='Last'
                        )
                    )

                else:
                    number_to_be_bought = None

            else:
                raise ValueError('Ambiguous signal:', row['signal_type'])

            try:
                amount = df_trading_journal['Cash'].iloc[-1]
            except:
                amount = float_budget_in_usd

            if number_to_be_bought is None:
                number_to_be_bought = 0

            number_to_be_bought = round(number_to_be_bought, general_settings['rounding_decimal_places_for_security_quantities'])

            dataseries_trading_journal = execute_order(
                boolean_buy=boolean_buy,
                index=index,
                date=index[0],
                strategy_id=int_chosen_strategy,
                crypto_key=crypto_key,
                number_to_be_bought=number_to_be_bought,
                df_prices=df_prices,
                df_trading_journal=df_trading_journal,
                margin_loan_rate=margin_loan_rate,
                float_budget_in_usd=float_budget_in_usd,
                price=price,
                fees={
                    'absolute_fee_buy_order':dict_crypto_options['general']['absolute_fee_buy_order'],
                    'absolute_fee_sell_order':dict_crypto_options['general']['absolute_fee_sell_order'],
                    'percentage_buying_fees_and_spread':dict_crypto_options['general']['percentage_buying_fees_and_spread'],
                    'percentage_selling_fees_and_spread':dict_crypto_options['general']['percentage_selling_fees_and_spread']
                },
                display_options=display_options,
                constraints=constraints,
                general_settings=general_settings,
                boolean_allow_partially_filled_orders=strategy_hyperparameters['boolean_allow_partially_filled_orders']
            )

            df_trading_journal = df_trading_journal.append(
                dataseries_trading_journal,
                ignore_index=True
            )

    comments['constraints'] = constraints
    comments['general_settings'] = general_settings

    dict_return = {
        'df_trading_journal': df_trading_journal,
        'Strategy ID': '3',
        'Strategy label': 'X',
        'strategy_hyperparameters': strategy_hyperparameters,
        'comments': comments
    }

    save_dataframe_to_csv(
        df_trading_journal,
        'trading_journal',
        string_directory=display_options['string_results_directory'],
    )

    return dict_return

def execute_strategy_white_noise(
        df_prices,
        int_chosen_strategy,
        float_budget_in_usd,
        margin_loan_rate,
        list_trading_execution_delay_after_signal_in_hours,
        dict_crypto_options,
        minimum_expected_mispricing_trigger_in_percent,
        strategy_hyperparameters,
        display_options,
        constraints,
        general_settings,
        sell_at_the_end,
        comments,
        boolean_allow_shorting=False,
    ):
    """Executes buy and sell orders in alternating order.

    Has positive average exposure and is therefore not expected to yield a gross
    return of zero.
    """
    df_prices = df_prices.loc[(slice(strategy_hyperparameters['start_time'], strategy_hyperparameters['end_time']), strategy_hyperparameters['id']), :]

    df_trading_journal = initialize_trading_journal()

    usd_safety_buffer = 100

    # Single-asset only
    crypto_key = strategy_hyperparameters['id']

    for index, row in df_prices.iterrows():
        try:
            amount = df_trading_journal['Cash'].iloc[-1]
        except:
            amount = float_budget_in_usd

        if choice(['buy', 'sell']) == 'buy':
            # Determine the number of assets to be bought or sold
            try:
                float_budget_in_usd = df_trading_journal['Cash'].iloc[-1]
            except:
                pass

            number_to_be_bought = round(
                (
                    float_budget_in_usd - usd_safety_buffer
                ) / row['price'],
                general_settings['rounding_decimal_places_for_security_quantities']
            )

            if number_to_be_bought < 0:
                number_to_be_bought = 0

            boolean_buy = True

        else:
            # Determine the number of assets to be bought or sold
            try:
                number_to_be_bought = (-1) * df_trading_journal[
                    'Exposure (number)'
                ].iloc[-1]
            except:
                number_to_be_bought = 0

            boolean_buy = False

        dataseries_trading_journal = execute_order(
            df_prices=df_prices,
            df_trading_journal=df_trading_journal,
            boolean_buy=boolean_buy,
            index=index,
            date=index[0],
            strategy_id=int_chosen_strategy,
            crypto_key=crypto_key,
            number_to_be_bought=number_to_be_bought,
            margin_loan_rate=margin_loan_rate,
            float_budget_in_usd=float_budget_in_usd,
            price=row['price'],
            fees={
                'absolute_fee_buy_order':dict_crypto_options['general']['absolute_fee_buy_order'],
                'absolute_fee_sell_order':dict_crypto_options['general']['absolute_fee_sell_order'],
                'percentage_buying_fees_and_spread':dict_crypto_options['general']['percentage_buying_fees_and_spread'],
                'percentage_selling_fees_and_spread':dict_crypto_options['general']['percentage_selling_fees_and_spread']
            },
            display_options=display_options,
            constraints=constraints,
            general_settings=general_settings,
            boolean_allow_partially_filled_orders=strategy_hyperparameters['boolean_allow_partially_filled_orders']
        )

        df_trading_journal = df_trading_journal.append(
            dataseries_trading_journal,
            ignore_index=True
        )

    dict_return = {
        'df_trading_journal': df_trading_journal,
        'Strategy ID': int_chosen_strategy,
        'Strategy label': 'White Noise',
        'strategy_hyperparameters': strategy_hyperparameters,
        'comments': ''
    }

    save_dataframe_to_csv(
        df_trading_journal,
        'trading_journal',
        string_directory=display_options['string_results_directory'],
    )

    return dict_return

def execute_strategy_ma_crossover(
        df_prices,
        int_chosen_strategy,
        float_budget_in_usd,
        margin_loan_rate,
        boolean_allow_shorting,
        list_trading_execution_delay_after_signal_in_hours,
        dict_crypto_options,
        general_settings,
        minimum_expected_mispricing_trigger_in_percent,
        strategy_hyperparameters,
        sell_at_the_end,
        file_path_with_signal_data,
        display_options,
        constraints,
        comments
    ):
    """Filters signals and executes all remaining signals."""
    df_prices['moving_average'] = df_prices.groupby(
        level='id'
    )['price'].transform(
        lambda x: round(
            x.rolling(
                window=strategy_hyperparameters['moving_average_window_in_days'],
                # on='datetime'
            ).mean(),
            general_settings['rounding_decimal_places']
        )
    )

    df_trading_journal = initialize_trading_journal()

    price = None

    current_time = strategy_hyperparameters['start_time']
    previous_time = Timestamp(current_time) - strategy_hyperparameters['frequency']

    times_to_loop_over = date_range(start=strategy_hyperparameters['start_time'], end=strategy_hyperparameters['end_time'], freq=strategy_hyperparameters['frequency']).to_series()
    for current_time in times_to_loop_over:
    #for time_elapsed in tqdm(range(strategy_hyperparameters['frequency'], ((strategy_hyperparameters['end_time'] - strategy_hyperparameters['start_time']) + strategy_hyperparameters['frequency'])), desc='Going through signals', unit='signal'):

        boolean_buy = None
        number_to_be_bought = 0

        # CONTINUE HERE
        previous_time = current_time - Timedelta(strategy_hyperparameters['frequency'])
        # This intermediate step is used to erase the frequency from the Timestamp
        previous_time = Timestamp(str(previous_time))

        old_row = df_prices.loc[(previous_time, strategy_hyperparameters['id']), : ]
        previous_ma = old_row['moving_average']
        old_price = old_row['price']

        new_ma = df_prices.loc[(current_time, strategy_hyperparameters['id']), 'moving_average']

        price = find_price(
            df_prices,
            desired_index=(current_time, strategy_hyperparameters['id']),
            boolean_allow_older_prices=False,
            boolean_allow_newer_prices=False,
            boolean_warnings=display_options['warning_no_price_during_execution']
        )

        if (previous_ma < old_price) and (new_ma > price):
            moving_average_crossover = 'Upside breach'
        elif (previous_ma > old_price) and (new_ma < price):
            moving_average_crossover = 'Downside breach'
        elif new_ma < price:
            moving_average_crossover = 'Above'
        elif new_ma > price:
            moving_average_crossover = 'Below'
        else:
            moving_average_crossover = None

        if moving_average_crossover == 'Upside breach' or moving_average_crossover == 'Above':
            boolean_buy = True
        elif moving_average_crossover == 'Downside breach' or moving_average_crossover == 'Below':
            boolean_buy = False
        else:
            boolean_buy = None

        if price is not None and price > 0:
            if boolean_buy:
                try:
                    float_budget_in_usd = df_trading_journal['Cash'].iloc[-1]
                except:
                    pass

                number_to_be_bought = round(
                    (
                        float_budget_in_usd
                        - constraints['minimum_cash']
                    ) / price,
                    general_settings['rounding_decimal_places_for_security_quantities']
                )

                if number_to_be_bought is not None:
                    number_to_be_bought = round(
                        number_to_be_bought * strategy_hyperparameters['maximum_relative_exposure_per_buy'],
                        general_settings['rounding_decimal_places_for_security_quantities']
                    )

            elif boolean_buy == False:
                if len(df_trading_journal) > 0:
                    number_to_be_bought = (-1) * round(
                        df_trading_journal.iloc[-1]['Dict of assets in portfolio'][strategy_hyperparameters['id']],
                        general_settings['rounding_decimal_places_for_security_quantities']
                    )
                else:
                    number_to_be_bought = 0
            else:
                number_to_be_bought = 0

            if number_to_be_bought != 0:
                dataseries_trading_journal = execute_order(
                    boolean_buy=boolean_buy,
                    index=current_time,
                    date=current_time,
                    strategy_id=int_chosen_strategy,
                    crypto_key=strategy_hyperparameters['id'],
                    number_to_be_bought=number_to_be_bought,
                    df_prices=df_prices,
                    df_trading_journal=df_trading_journal,
                    margin_loan_rate=margin_loan_rate,
                    float_budget_in_usd=float_budget_in_usd,
                    price=price,
                    fees={
                        'absolute_fee_buy_order': dict_crypto_options['general']['absolute_fee_buy_order'],
                        'absolute_fee_sell_order': dict_crypto_options['general']['absolute_fee_sell_order'],
                        'percentage_buying_fees_and_spread': dict_crypto_options['general']['percentage_buying_fees_and_spread'],
                        'percentage_selling_fees_and_spread': dict_crypto_options['general']['percentage_selling_fees_and_spread']
                    },
                    display_options=display_options,
                    constraints=constraints,
                    general_settings=general_settings,
                    boolean_allow_partially_filled_orders=strategy_hyperparameters['boolean_allow_partially_filled_orders']
                )

                df_trading_journal = df_trading_journal.append(
                    dataseries_trading_journal,
                    ignore_index=True
                )

    dict_return = {
        'df_trading_journal': df_trading_journal,
        'Strategy ID': '4',
        'Strategy label': 'Moving Average Crossover',
        'strategy_hyperparameters': strategy_hyperparameters,
        'comments': comments
    }

    save_dataframe_to_csv(
        df_trading_journal,
        'trading_journal',
        string_directory=display_options['string_results_directory'],
    )

    return dict_return

def test_strategies(
        df_prices,
        int_chosen_strategy,
        float_budget_in_usd,
        margin_loan_rate,
        boolean_allow_shorting,
        list_trading_execution_delay_after_signal_in_hours,
        dict_crypto_options,
        minimum_expected_mispricing_trigger_in_percent,
        strategy_hyperparameters,
        sell_at_the_end,
        benchmark_data_specifications,
        display_options,
        constraints,
        general_settings,
        start_time,
        file_path_with_signal_data=None,
        comments={}
    ):
    """Calls user-defined strategy.

    Chooses the correct strategy (as per user input) and returns
    the performance metrics and the trading journal of that strategy.
    """

    if int_chosen_strategy == 1:
        dict_execution_results = execute_strategy_white_noise(
            df_prices=df_prices,
            float_budget_in_usd=float_budget_in_usd,
            margin_loan_rate=margin_loan_rate,
            boolean_allow_shorting=boolean_allow_shorting,
            list_trading_execution_delay_after_signal_in_hours=list_trading_execution_delay_after_signal_in_hours,
            int_chosen_strategy=int_chosen_strategy,
            dict_crypto_options=dict_crypto_options,
            minimum_expected_mispricing_trigger_in_percent=minimum_expected_mispricing_trigger_in_percent,
            strategy_hyperparameters=strategy_hyperparameters,
            display_options=display_options,
            constraints=constraints,
            general_settings=general_settings,
            sell_at_the_end=sell_at_the_end,
            comments=comments
        )
    elif int_chosen_strategy == 2:
        raise NotImplementedError('Strategy 2 is not implemented.')
    elif int_chosen_strategy == 3:
        dict_execution_results = execute_strategy_multi_asset(
            df_prices=df_prices,
            file_path_with_signal_data=file_path_with_signal_data,
            float_budget_in_usd=float_budget_in_usd,
            int_chosen_strategy=int_chosen_strategy,
            margin_loan_rate=margin_loan_rate,
            boolean_allow_shorting=boolean_allow_shorting,
            list_trading_execution_delay_after_signal_in_hours=list_trading_execution_delay_after_signal_in_hours,
            dict_crypto_options=dict_crypto_options,
            minimum_expected_mispricing_trigger_in_percent=minimum_expected_mispricing_trigger_in_percent,
            strategy_hyperparameters=strategy_hyperparameters,
            sell_at_the_end=sell_at_the_end,
            display_options=display_options,
            constraints=constraints,
            general_settings=general_settings,
            comments=comments
        )
    elif int_chosen_strategy == 4:
        dict_execution_results = execute_strategy_ma_crossover(
            df_prices=df_prices,
            file_path_with_signal_data=file_path_with_signal_data,
            int_chosen_strategy=int_chosen_strategy,
            float_budget_in_usd=float_budget_in_usd,
            margin_loan_rate=margin_loan_rate,
            boolean_allow_shorting=boolean_allow_shorting,
            list_trading_execution_delay_after_signal_in_hours=list_trading_execution_delay_after_signal_in_hours,
            dict_crypto_options=dict_crypto_options,
            minimum_expected_mispricing_trigger_in_percent=minimum_expected_mispricing_trigger_in_percent,
            strategy_hyperparameters=strategy_hyperparameters,
            sell_at_the_end=sell_at_the_end,
            display_options=display_options,
            constraints=constraints,
            general_settings=general_settings,
            comments=comments
        )

    save_dataframe_to_csv(
        dict_execution_results['df_trading_journal'],
        string_name='trading_journal',
        string_directory=display_options['string_results_directory'],
    )

    df_performance = evaluate_performance(
        df_prices=df_prices,
        dict_execution_results=dict_execution_results,
        float_budget_in_usd=float_budget_in_usd,
        benchmark_data_specifications=benchmark_data_specifications,
        strategy_hyperparameters=strategy_hyperparameters,
        display_options=display_options,
        constraints=constraints,
        general_settings=general_settings,
        start_time=start_time,
    )

    return [df_performance, dict_execution_results['df_trading_journal']]



"""Provides useful general functions for other modules."""

# For managing dates
from datetime import datetime, timedelta

# For mathematical operations
from math import isnan

# For managing tables properly
from pandas import DataFrame, Timedelta

# For deep-copied dictionaries
from copy import deepcopy


def find_dataframe_value_with_keywords(
        df,
        search_term_1,
        search_column_name_1,
        search_term_2=None,
        search_column_name_2=None,
        output_column_name=None,
        first_last_or_all_elements='All'
    ):
    """
    Finds a match from a Pandas DataFrame from one or two column-based criteria.
    Output may be specified as being the first match, the last match, or all
    matches.

    Returns None if no match is found.
    """
    df = df.reset_index()

    result = None

    if search_term_2 is None:
        corresponding_index_list = df.index[
            df[search_column_name_1] == search_term_1
        ].tolist()

        result = df[df.index.isin(corresponding_index_list)]

    else:
        df = df[df[search_column_name_1].isin([search_term_1])]

        result = find_dataframe_value_with_keywords(
            df,
            search_term_1=search_term_2,
            search_column_name_1=search_column_name_2
        )

    if isinstance(result, DataFrame):
        result = result.reset_index()

    if first_last_or_all_elements == 'First':
        try:
            result = result.iloc[0]
        except IndexError:
            raise ValueError('Search term', search_term_1, 'not found.')
        if output_column_name is not None:
            result = result[output_column_name]

    elif first_last_or_all_elements == 'Last':
        try:
            result = result.iloc[-1]
        except IndexError:
            raise ValueError('Search term', search_term_1, 'not found.')
        if output_column_name is not None:
            result = result[output_column_name]

    elif first_last_or_all_elements == 'All':
        if output_column_name is not None:
            output = []
            for index, row in result.iterrows():
                output.append(row[output_column_name])
            result = output

    else:
        raise ValueError('The method find_dataframe_value_with_keywords() received an invalid input for argument first_last_or_all_elements.')

    try:
        if len(result) == 0:
            result = None

    except:
        pass

    return result

def find_price(
        df_prices,
        desired_index,
        boolean_allow_older_prices=False,
        boolean_allow_newer_prices=False,
        boolean_warnings=True,
        boolean_errors=False,
        time_to_allow_to_go_back=Timedelta(days=30),
        time_to_allow_to_go_to_the_future=Timedelta(days=30)
    ):
    """
    This function finds the price of a security at a given time from a
    dataframe. If no price is available, it can search for the next older
    price while issuing a warning. If there is also no next older price, it can
    search for the next newer price while issuing a warning. If there is no
    price at all, it throws an error.

    Substituting current prices by older or newer prices should only be used for
    the portfolio evaluation of existing assets, not for buying or selling
    existing assets. The impact of actually simulating orders with nonoptimal
    (= approximated via this function) prices may lead to errors that are too
    large to accept. Using this function for intermediate portfolio valuation,
    however, is expected to have a limited impact on overall performance.
    """
    price = None

    try:
        try:
            price = df_prices.loc[desired_index, 'price'][0]
        except IndexError:
            price = df_prices.loc[desired_index, 'price']
    except KeyError:
        try:
            if boolean_allow_older_prices:
                matched_row = df_prices.loc[
                    (
                        slice(
                            desired_index[0]-time_to_allow_to_go_back,
                            desired_index[0]
                        ),
                        desired_index[1]
                    )
                ][-1]
                price = matched_row.price

                if price is not None:
                    if boolean_warnings:
                        print(f'Find price warning: An older date was found for {desired_index[1]}. Instead of the time {desired_index[0]} the data only offers a price for {matched_row.index[0]}. Price: {str(price)}')
                    if boolean_errors:
                        raise ValueError('No price found.')

        except KeyError:
            pass

        # The search for newer prices is skipped if an older price was already found. Older prices are assumed to be strictly better than newer prices.
        try:
            if (price == None) and boolean_allow_newer_prices:
                matched_row = df_prices.loc[(slice(desired_index[0], desired_index[0] + time_to_allow_to_go_to_the_future), desired_index[1])][0]
                price = matched_row.price

                if price is not None:
                    if boolean_warnings:
                        print(f'Find price warning: A newer date was found for {desired_index[1]}. Instead of the time {str(desired_index[0])}, the data only offers a price for {matched_row.index[0]}. Price: {str(price)}')
                    if boolean_errors:
                        raise ValueError('No price found.')

        except KeyError:
            pass

    if price is not None:
        if isnan(price):
            price = None

    if boolean_warnings:
        if price is None and boolean_allow_older_prices and boolean_allow_newer_prices:
            print(f'Find price warning: No price was found for {desired_index[1]}. Neither for {str(desired_index[0])}, nor for any other point in time.')
        elif price is None and boolean_allow_newer_prices:
            print(f'Find price warning: No price was found for {desired_index[1]}. Neither for {str(desired_index[0])}, nor for an earlier point in time.')
        elif price is None and boolean_allow_older_prices:
            print(f'Find price warning: No price was found for {desired_index[1]}. Neither for {str(desired_index[0])}, nor for an older point in time.')
        elif price is None:
            print(f'Find price warning: No price was found for {desired_index[1]} at date {str(desired_index[0])}')

    price = check_for_data_error_and_return_alternative_price(
        desired_index,
        price,
        boolean_warnings=True
    )

    return price

def add_time_column_to_dataframe_from_string(
        df,
        string_column_name,
        output_column_name=None
    ):
    """Adds a datetime object to the DataFrame, created from a string date."""
    if output_column_name is None:
        output_column_name = string_column_name

    list_times = []
    for index, row in df.iterrows():
        current_date = string_to_datetime(row[string_column_name])
        list_times.append(current_date)

    df[output_column_name] = list_times

    return df

def alternative_date_finder(
        df,
        initial_date,
        date_column_name,
        maximum_deviation_in_days,
        boolean_look_for_older_dates,
        boolean_warnings
    ):
    """Finds the next date at or around a given date in a DataFrame."""
    date_found = False
    new_date = initial_date
    for no_of_days_shift in range(0, maximum_deviation_in_days + 1):
        if (new_date == df[date_column_name]).any():
            date_found = new_date
            break
        else:
            if boolean_look_for_older_dates:
                try:
                    new_date = initial_date - timedelta(days=no_of_days_shift)
                except:
                    new_date = initial_date - timedelta(days=no_of_days_shift)
            else:
                try:
                    new_date = initial_date + timedelta(days=no_of_days_shift)
                except:
                    new_date = initial_date + timedelta(days=no_of_days_shift)

    if date_found is False:
        raise Exception(f'No alternative date found for the given date {initial_date}')
    else:
        if boolean_warnings:
            if initial_date == new_date:
                print(f'{initial_date} is covered.')
            else:
                print(f'An alternative date was found. {new_date} instead of {initial_date}')

    return new_date

def string_to_datetime(string_object):
    if isinstance(string_object, datetime):
        print(f'Alternative date finder warning: Was already a datetime.datetime object. Input was: {string_object}')
        return string_object
    else:
        try:
            current_year = int(string_object[:4])
            current_month = int(string_object[5:7])
            current_day = int(string_object[8:10])
            current_date = datetime(
                year=current_year,
                month=current_month,
                day=current_day,
                hour=0,
                minute=0,
                second=0,
                microsecond=0
            )
        except ValueError:
            print('Please make sure to use the international date format for the date column: yyyy-mm-dd.')

        return current_date

    raise NotImplementedError('Unexpected error. Please contact the maintainer of this code.')

def datetime_to_string(datetime_object):
    if isinstance(datetime_object, str):
        string_object = datetime_object
    else:
        string_object = str(datetime_object)[:10]

    return string_object

def check_for_data_error_and_return_alternative_price(
        desired_index,
        price,
        boolean_warnings=True
    ):
    """Manually maintained function that contains known data issues."""
    boolean_change = False

    if desired_index == (datetime(year=2018, month=5, day=4), '0x4672bad527107471cb5067a887f4656d585a8a31'):
        old_price = price
        price = 0.0057
        boolean_change = True

    if boolean_warnings and boolean_change:
        print(f'Check for data error warning: There was a known error with this data point. An alternative price is used id: {desired_index[1]}, time: {desired_index[0]}, old price: {old_price}, new price: {price}')

    return price

def calculate_portfolio_value(
        df_prices,
        dict_of_assets_in_portfolio,
        time,
        constraints,
        display_options,
        rounding_accuracy,
        cash_value=0.0,
        boolean_gross_exposure=False
    ):
    """Calculates the value of a portfolio, excluding cash."""

    if boolean_gross_exposure:
        if cash_value == 0.0:
            dict_of_assets_in_portfolio = deepcopy(dict_of_assets_in_portfolio)
            for asset in dict_of_assets_in_portfolio:
                dict_of_assets_in_portfolio[asset] = abs(
                    dict_of_assets_in_portfolio[asset]
                )
        else:
            raise ValueError('When calculating gross exposure, the cash argument must be zero.')

    total_exposure = 0

    for asset in dict_of_assets_in_portfolio:
        if dict_of_assets_in_portfolio[asset] != 0:
            # Take current price for that asset to calculate exposure in $/€

            best_price = find_price(
                df_prices=df_prices,
                desired_index=(time, asset),
                boolean_allow_older_prices=True,
                boolean_allow_newer_prices=False,
                boolean_warnings=display_options['warning_no_price_for_intermediate_valuation']
            )

            if best_price is None:
                best_price = 0
            elif not best_price > 0:
                best_price = 0

            result = dict_of_assets_in_portfolio[asset] * best_price

            total_exposure = total_exposure + result

            if (constraints['boolean_allow_shortselling'] is False) and (dict_of_assets_in_portfolio[asset] < 0):
                raise NotImplementedError(f'Exposure cannot be smaller than zero. \nPortfolio: {dict_of_assets_in_portfolio}')

    total_exposure = round(total_exposure + cash_value, rounding_accuracy)

    return total_exposure

def calculate_relative_gross_exposure(
        df_prices,
        dict_of_assets_in_portfolio,
        time,
        display_options,
        constraints,
        rounding_accuracy,
        cash_value,
        general_settings
    ):

    absolute_gross_exposure_before_potential_trade = calculate_portfolio_value(
        df_prices=df_prices,
        dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
        time=time,
        constraints=constraints,
        display_options=display_options,
        rounding_accuracy=rounding_accuracy,
        boolean_gross_exposure=True
    )

    portfolio_value = calculate_portfolio_value(
        df_prices=df_prices,
        dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
        time=time,
        constraints=constraints,
        display_options=display_options,
        rounding_accuracy=rounding_accuracy,
        cash_value=cash_value,
        boolean_gross_exposure=False
    )

    relative_gross_exposure = absolute_gross_exposure_before_potential_trade / portfolio_value

    return relative_gross_exposure



"""This module contains functions that load, save, and manipulate data."""

# For counting the number of result files in the "results" folder to allow for
# consistent file versioning when saving the results as a csv file.
from os import listdir, path

# For managing tables properly
from pandas import read_csv, merge
import pandas


def load_data(
        file_path_with_price_data,
        strategy_hyperparameters,
        file_path_with_token_data=None,
        name_of_foreign_key_in_price_data_table=None,
        name_of_foreign_key_in_token_metadata_table=None
    ):
    """Loads price data and metadata.

    This function loads prices and token metadata (token data). It also uses the
    date as an index.
    """

    # Handles different csv formats and handles program calls from root and from
    # /backtesting
    try:
        df_prices = read_csv(
            file_path_with_price_data,
            sep=',',
            usecols=['datetime', 'token_itin', 'price'],
            parse_dates=['datetime'],
            infer_datetime_format=True
        )
    except (pandas.errors.ParserError, FileNotFoundError, ValueError):
        df_prices = read_csv(
            file_path_with_price_data,
            sep=';',
            usecols=['datetime', 'token_itin', 'price'],
            parse_dates=['datetime'],
            infer_datetime_format=True
        )
    # Enrich data with metadata
    if file_path_with_token_data is not None:
        try:
            df_token_metadata = read_csv(
                file_path_with_token_data,
                sep=';',
                usecols=['token_itin', 'token_address_eth']
            )
        except ValueError:
            df_token_metadata = read_csv(
                file_path_with_token_data,
                sep=',',
                usecols=['token_itin', 'token_address_eth']
            )

        df_prices = merge_data(
            df_prices=df_prices,
            df_token_metadata=df_token_metadata,
            name_of_foreign_key_in_price_data_table=name_of_foreign_key_in_price_data_table,
            name_of_foreign_key_in_token_metadata_table=name_of_foreign_key_in_token_metadata_table
        )

    df_prices.rename(
        columns={
            strategy_hyperparameters['prices_table_id_column_name']: 'id'
        },
        inplace=True
    )

    df_prices.set_index(['datetime', 'id'], inplace=True)

    df_prices.sort_index(level=['datetime', 'id'], ascending=[1, 1], inplace=True)

    if not df_prices.index.is_lexsorted():
        raise ValueError('df_prices is not lexsorted.')

    return df_prices

def merge_data(
        df_prices,
        df_token_metadata,
        name_of_foreign_key_in_price_data_table,
        name_of_foreign_key_in_token_metadata_table
    ):
    """Enriches price data with metadata about the tokens."""

    df_prices_enriched_with_metadata = merge(
        df_prices,
        df_token_metadata,
        left_on=name_of_foreign_key_in_price_data_table,
        right_on=name_of_foreign_key_in_token_metadata_table
    )

    return df_prices_enriched_with_metadata

def prepare_data(
        file_path_with_price_data,
        strategy_hyperparameters,
        file_path_with_token_data=None,
        name_of_foreign_key_in_price_data_table=None,
        name_of_foreign_key_in_token_metadata_table=None
    ):
    """Loads prices and token data and joins those into a pandas DataFrame."""

    df_prices = load_data(
        file_path_with_price_data=file_path_with_price_data,
        file_path_with_token_data=file_path_with_token_data,
        strategy_hyperparameters=strategy_hyperparameters,
        name_of_foreign_key_in_price_data_table=name_of_foreign_key_in_price_data_table,
        name_of_foreign_key_in_token_metadata_table=name_of_foreign_key_in_token_metadata_table
    )

    return df_prices

def save_dataframe_to_csv(
        df_prices,
        string_name,
        string_directory,
    ):
    """Saves a pandas Dataframe to csv without overwriting existing files."""
    result_no = len(
        [name for name in listdir(string_directory) if path.isfile(
            path.join(
                string_directory,
                name
            )
        )]
    ) / 2


    number_of_result_files_plus_1 = 1 + int(result_no)
    df_prices.to_csv(
        string_directory + '/' + string_name + '_' + str(number_of_result_files_plus_1) + '.csv'
    )



"""Provides functions for performance evaluations of the trading journal.

This module is more high-level than _3_individual_metrics."""

# For deep-copied dictionaries
from copy import deepcopy

# For managing tables properly
from pandas import DataFrame

# For plotting return curve
import matplotlib.pyplot as plt
from matplotlib.dates import MonthLocator
from matplotlib.dates import DateFormatter
import matplotlib.ticker as tickercalculate_alpha

# For managing dates
from datetime import datetime, timedelta

# For counting the number of result files in the "results" folder to allow for
# consistent file versioning when saving the results as a CSV file.
import os

from _1_data_preparation import save_dataframe_to_csv
from _3_individual_metrics import calculate_volatility, calculate_roi, \
    calculate_sharpe_ratio, calculate_maximum_drawdown, \
    calculate_transaction_cost, calculate_beta, calculate_alpha
from _helper_functions import find_dataframe_value_with_keywords, find_price, \
    alternative_date_finder, calculate_portfolio_value, datetime_to_string, \
    add_time_column_to_dataframe_from_string


def calculate_returns_single(
        previous_trading_journal_row,
        current_trading_journal_row,
        df_prices,
        strategy_hyperparameters,
        display_options,
        general_settings,
        constraints
    ):
    """Returns a list of dicts with portfolio return data for a given frequency.

    Dict fields:
        'timestamp' (datetime.datetime), 'portfolio_value' (float), 'return'
        (float), 'relative_return' (float), 'dict_of_assets_in_portfolio'
        (dict with itins as keys and integer with the number of pieces held of
        this asset as as values)

    The reasoning behind the return calculation and the frequency handling is
    described using an exemplary time series. The frequency is minutes here, so
    frequency_in_seconds=60. The first column represents the executions
    (previous_trading_journal_row and current_trading_journal_row), the second
    column represents the frequency.

    Trades                      Frequency increments

    -No trade-                  Minute 1
    Trade 1                     Minute 2
    -No trade-                  Minute 3
    -No trade-                  Minute 4
    Trade 2                     Minute 5

    In the example above, the return calculation would work as follows: The
    function would receive DataFrame rows for Trade 1 and Trade 2. It would run
    the return calculation loop for Minute 2-3, Minute 3-4, and Minute 4-5.
    Thus, the function would return a list of three dicts. The returns for
    Minute 1-2 were already calculated in an earlier function call (Trade 0-1).
    It is crucial that the datetime.datetime objects that are contained in
    Trade 1 and Trade 2 have the same frequency as the frequency that is passed
    as an argument. Otherwise, there can be missing entries in the aggregated
    return DataFrame.

    datetime.datetime objects assume 'there are exactly 3600*24 seconds in every
    day'. https://docs.python.org/2/library/datetime.html#datetime-objects
    """
    list_of_dict_returns = []

    dict_of_assets_in_portfolio = previous_trading_journal_row['Dict of assets in portfolio']
    copy_dict_of_assets_in_portfolio = deepcopy(dict_of_assets_in_portfolio)

    cash = previous_trading_journal_row['Cash']

    datetime_counter = previous_trading_journal_row['datetime'].to_pydatetime()

    while (
            datetime_counter + strategy_hyperparameters['frequency']
    ) <= (
            current_trading_journal_row['datetime'].to_pydatetime()
    ):
        returns = {'dict_of_assets_in_portfolio': copy_dict_of_assets_in_portfolio}

        # The counter is purposefully incremented at the beginning
        datetime_counter += strategy_hyperparameters['frequency']

        returns['datetime'] = datetime_counter

        previous_portfolio_value = calculate_portfolio_value(
            df_prices=df_prices,
            dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
            time=datetime_counter - strategy_hyperparameters['frequency'],
            cash_value=cash,
            display_options=display_options,
            constraints=constraints,
            rounding_accuracy=general_settings['rounding_decimal_places']
        )

        returns['portfolio_value'] = calculate_portfolio_value(
            df_prices=df_prices,
            dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
            time=datetime_counter,
            cash_value=cash,
            display_options=display_options,
            constraints=constraints,
            rounding_accuracy=general_settings['rounding_decimal_places']
        )

        returns['dict_of_assets_in_portfolio'] = deepcopy(copy_dict_of_assets_in_portfolio)

        try:
            returns['portfolio_return'] = round(
                returns['portfolio_value'] - previous_portfolio_value,
                general_settings['rounding_decimal_places']
            )
            try:
                returns['relative_portfolio_return'] = round(
                    returns['portfolio_return'] / previous_portfolio_value,
                    general_settings['rounding_decimal_places']
                )
            except ZeroDivisionError:
                raise ZeroDivisionError(
                    """
                        This is a rather unlikely scenario, therefore an error
                        was raised. The previous portfolio value is zero. This
                        cannot be.
                    """ +
                    '\nportfolio_value: ' + str(
                        returns['portfolio_value']
                    ) +
                    '\nportfolio_return: ' + str(
                        returns['portfolio_return']
                    )
                )
        except IndexError:
            returns['portfolio_return'] = round(
                returns['portfolio_value'] - df_trading_journal['Cash before'].iloc[0],
                general_settings['rounding_decimal_places']
            )
            returns['relative_portfolio_return'] = round(
                returns['portfolio_return'] / df_trading_journal['Cash before'].iloc[0],
                general_settings['rounding_decimal_places']
            )

        returns = deepcopy(returns)

        list_of_dict_returns.append(returns)

    return list_of_dict_returns

def calculate_returns_batch(
        df_trading_journal,
        df_prices,
        strategy_hyperparameters,
        display_options,
        general_settings,
        constraints
    ):
    """Calculates returns of a portfolio from a trading journal.

    Any frequency between miliseconds and infinity can be used (hourly, daily
    weekly, etc.).
    """

    if len(df_trading_journal) < 1:
        raise ValueError(f'There were no trades. The provided trading journal has {str(len(df_trading_journal))} trades.')

    df_returns = DataFrame(
        columns=[
            'datetime',
            'portfolio_value',
            'portfolio_return',
            'relative_portfolio_return',
            'dict_of_assets_in_portfolio'
        ]
    )

    df_returns.set_index(
        keys=['datetime'],
        inplace=True
    )

    first_date = df_trading_journal['datetime'].iloc[0]
    last_date = df_trading_journal['datetime'].iloc[-1]

    if df_trading_journal['Cash before'].iloc[0] is None:
        raise ValueError('initial_budget should not be None.')

    current_date = first_date
    for index, row in df_trading_journal.iterrows():
        if index > 0:
            previous_trading_journal_row = df_trading_journal.loc[index - 1]
            current_trading_journal_row = df_trading_journal.loc[index]

            list_of_dict_returns = calculate_returns_single(
                previous_trading_journal_row=previous_trading_journal_row,
                current_trading_journal_row=current_trading_journal_row,
                df_prices=df_prices,
                strategy_hyperparameters=strategy_hyperparameters,
                display_options=display_options,
                general_settings=general_settings,
                constraints=constraints
            )

            for dict_returns in list_of_dict_returns:
                df_returns.loc[dict_returns['datetime']] = {
                    'portfolio_value': dict_returns['portfolio_value'],
                    'portfolio_return': dict_returns['portfolio_return'],
                    'relative_portfolio_return': dict_returns['relative_portfolio_return'],
                    'dict_of_assets_in_portfolio': dict_returns['dict_of_assets_in_portfolio']
                }
        else:
            pass

    assert len(df_returns) > 0

    return df_returns

def calculate_daily_returns_from_benchmark(
        first_date,
        last_date,
        df_prices,
        benchmark_id,
        display_options,
        strategy_hyperparameters,
        general_settings,
        constraints
    ):
    """Calculates daily returns of a benchmark."""
    df_daily_returns = DataFrame(
        columns=['benchmark_datetime', 'benchmark_dict_of_assets_in_portfolio', 'benchmark_portfolio_value', 'benchmark_return', 'benchmark_relative_return']
    )

    df_daily_returns.set_index(['benchmark_datetime'], inplace=True)

    initial_budget = find_price(
        df_prices=df_prices,
        desired_index=(first_date, slice(None)),
        boolean_allow_older_prices=False,
        boolean_allow_newer_prices=False,
        boolean_warnings=True,
        boolean_errors=True
    )

    if initial_budget is None:
        raise ValueError(f'initial_budget should not be None. Benchmark ID: {benchmark_id}. No price found for {first_date} in \n{df_prices}.')

    current_date = first_date
    for days_elapsed in range(1, ((last_date - first_date).days + 1)):
        current_date = first_date + timedelta(days=days_elapsed)
        # previous_date = current_date - timedelta(days=1)

        dict_of_assets_in_portfolio = {
            str(benchmark_id): 1.0
        }

        portfolio_value = calculate_portfolio_value(
            df_prices=df_prices,
            dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
            time=current_date,
            cash_value=0,
            rounding_accuracy=general_settings['rounding_decimal_places'],
            display_options=display_options,
            constraints=constraints
        )

        try:
            portfolio_return = round(
                portfolio_value - df_daily_returns['benchmark_portfolio_value'].iloc[-1],
                2
            )
            try:
                relative_portfolio_return = round(
                    portfolio_return / df_daily_returns['benchmark_portfolio_value'].iloc[-1],
                    2
                )
            except ZeroDivisionError:
                raise ZeroDivisionError(
                    """
                        This is a rather unlikely scenario, therefore an error
                        was raised. The previous portfolio value is zero. This
                        cannot be.
                    """ +
                    '\nprevious portfolio value: ' + str(
                        df_daily_returns['benchmark_portfolio_value'].iloc[-1]
                    ) +
                    '\nportfolio_value: ' + str(
                        portfolio_value
                    ) +
                    '\nportfolio_return: ' + str(
                        portfolio_return
                    )
                )
        except IndexError:
            portfolio_return = round(
                portfolio_value - initial_budget,
                2
            )
            relative_portfolio_return = round(
                portfolio_return / initial_budget,
                2
            )

        copy_dict_of_assets_in_portfolio = deepcopy(dict_of_assets_in_portfolio)

        df_daily_returns.loc[current_date] = {
            'benchmark_portfolio_value': portfolio_value,
            'benchmark_return': portfolio_return,
            'benchmark_relative_return': relative_portfolio_return,
            'benchmark_dict_of_assets_in_portfolio': copy_dict_of_assets_in_portfolio
        }

    assert len(df_daily_returns) > 0

    return df_daily_returns

def initialize_performance_overview():
    """Initializes a pandas DataFrame that serves as a performance overview.

    Initialization is important for determining the column order.
    """
    df_performance = DataFrame(columns=[
        'Strategy metadata --->',
        'Strategy ID',
        'Strategy label',
        'Trading info --->',
        'Begin time of tested interval',
        'End time of tested interval',
        'Duration of the tested interval',
        'Duration of the tested interval (in days)',
        'Average cash',
        'Average ticket size',
        'Number of trades',
        'Number of unique assets traded',
        'Total transaction cost',
        'Return metrics --->',
        'USD annualized ROI (from first to last trade)',
        'Cryptocurrency annualized ROI delta (from first to last trade)',
        'Ending benchmark value (first to last trade)',
        'Initial budget',
        'Ending portfolio value',
        'Risk metrics --->',
        'Holding period volatility',
        'Annual volatility',
        'Monthly volatility',
        'Weekly volatility',
        'Beta relative to benchmark',
        'Maximum drawdown',
        'Maximum drawdown duration',
        'Maximum drawdown peak date',
        'Maximum drawdown trough date',
        'Other metrics --->',
        'Alpha',
        'Sharpe ratio (holding period)',
        'Sharpe ratio (yearly)',
        'Beginning benchmark value (first to last trade)',
        'Other info --->',
        'Start time',
        'End time',
        'Parameter 1',
        'Parameter 2',
        'Comments',
        'Benchmark return metrics --->',
        'Benchmark USD annualized ROI (from first to last trade)',
        'Benchmark cryptocurrency annualized ROI delta (from first to last trade)',
        'Benchmark ending benchmark value (first to last trade)',
        'Benchmark initial budget',
        'Benchmark ending portfolio value',
        'Benchmark risk metrics --->',
        'Benchmark holding period volatility',
        'Benchmark annual volatility',
        'Benchmark monthly volatility',
        'Benchmark weekly volatility',
        'Benchmark Beta relative to benchmark',
        'Benchmark maximum drawdown',
        'Benchmark maximum drawdown duration',
        'Benchmark maximum drawdown peak date',
        'Benchmark maximum drawdown trough date',
        'Benchmark other metrics --->',
        'Benchmark Sharpe ratio (holding period)',
        'Benchmark Sharpe ratio (yearly)',
    ])

    return df_performance

def evaluate_performance(
        df_prices,
        dict_execution_results,
        float_budget_in_usd,
        benchmark_data_specifications,
        strategy_hyperparameters,
        display_options,
        general_settings,
        constraints,
        start_time
    ):
    """Evaluates the performance of a trading journal.

    Common metrics to evaluate trading strategies are used.
    """

    df_performance = initialize_performance_overview()

    df_trading_journal = dict_execution_results['df_trading_journal']

    try:
        df_daily_returns = calculate_returns_batch(
            df_trading_journal=df_trading_journal,
            df_prices=df_prices,
            strategy_hyperparameters=strategy_hyperparameters,
            display_options=display_options,
            general_settings=general_settings,
            constraints=constraints
        )
    except ValueError:
        print(f'Warning: No trades were executed with the given paremters: {strategy_hyperparameters}')
        return None

    if len(df_trading_journal) < 1:
        df_performance = df_performance.append(
            {
                'Strategy ID': None
            },
            ignore_index=True
        )

    else:
        begin_time_of_tested_interval = df_trading_journal[
            'datetime'
        ].iloc[0]
        end_time_of_tested_interval = dict_execution_results[
            'df_trading_journal'
        ]['datetime'].iloc[-1]

        # Ethereum does not have an eth_address and needs to be filtered usint the ITIN.
        try:
            df_benchmark = df_prices[df_prices['token_itin'] == benchmark_data_specifications['benchmark_key']]
        except:
            df_benchmark = df_prices.loc[(slice(None), benchmark_data_specifications['benchmark_key']), : ]

        df_daily_benchmark_returns = calculate_daily_returns_from_benchmark(
            first_date=begin_time_of_tested_interval,
            last_date=end_time_of_tested_interval,
            df_prices=df_benchmark,
            benchmark_id=benchmark_data_specifications['benchmark_key'],
            display_options=display_options,
            general_settings=general_settings,
            strategy_hyperparameters=strategy_hyperparameters,
            constraints=constraints
        )

        df_daily_returns['benchmark_portfolio_value'] = df_daily_benchmark_returns['benchmark_portfolio_value']
        df_daily_returns['benchmark_portfolio_value_normalized'] = df_daily_returns['benchmark_portfolio_value'] / df_daily_returns['benchmark_portfolio_value'][0]
        df_daily_returns['portfolio_value_normalized'] = df_daily_returns['portfolio_value'] / df_daily_returns['portfolio_value'][0]

        df_daily_returns['benchmark_portfolio_value'] = df_daily_benchmark_returns['benchmark_portfolio_value']
        df_daily_returns['benchmark_return'] = df_daily_benchmark_returns['benchmark_return']
        df_daily_returns['benchmark_relative_return'] =  df_daily_benchmark_returns['benchmark_relative_return']
        df_daily_returns['benchmark_dict_of_assets_in_portfolio'] =  df_daily_benchmark_returns['benchmark_dict_of_assets_in_portfolio']

        save_dataframe_to_csv(
            df_daily_returns,
            string_name='df_daily_returns',
            string_directory=display_options['string_results_directory'],
        )

        beginning_benchmark_value = find_dataframe_value_with_keywords(
            df_benchmark,
            search_term_1=df_trading_journal['datetime'].iloc[0],
            search_column_name_1='datetime',
            search_term_2=None,
            search_column_name_2=None,
            output_column_name='price',
            first_last_or_all_elements='First'
        )

        ending_benchmark_value = find_dataframe_value_with_keywords(
            df_benchmark,
            search_term_1=df_trading_journal['datetime'].iloc[-1],
            search_column_name_1='datetime',
            search_term_2=None,
            search_column_name_2=None,
            output_column_name='price',
            first_last_or_all_elements='First'
        )

        dict_roi_results = calculate_roi(
            df_trading_journal=df_trading_journal,
            float_budget_in_usd=float_budget_in_usd,
            df_benchmark=df_benchmark
        )

        # Average ticket size
        dict_execution_results['df_trading_journal']['Value bought absolute'] = dict_execution_results['df_trading_journal']['Value bought'].abs()
        average_ticket_size_absolute_value = round(
            dict_execution_results['df_trading_journal']['Value bought absolute'].mean(),
            2
        )

        maximum_drawdown, maximum_drawdown_duration, maximum_drawdown_peak_date, maximum_drawdown_trough_date = calculate_maximum_drawdown(
            df_daily_returns=df_daily_returns,
            column_with_portfolio_values='portfolio_value'
        )

        benchmark_maximum_drawdown, benchmark_maximum_drawdown_duration, benchmark_maximum_drawdown_peak_date, benchmark_maximum_drawdown_trough_date = calculate_maximum_drawdown(
            df_daily_returns=df_daily_returns,
            column_with_portfolio_values='benchmark_portfolio_value'
        )

        beta = calculate_beta(
            df_daily_returns,
            df_daily_benchmark_returns
        )

        df_performance = df_performance.append(
            {
                'Strategy ID': dict_execution_results['Strategy ID'],
                'Strategy label': dict_execution_results['Strategy label'],
                'Trading info --->': 'Trading info --->',
                'Begin time of tested interval': begin_time_of_tested_interval,
                'End time of tested interval': end_time_of_tested_interval,
                'Duration of the tested interval': end_time_of_tested_interval - begin_time_of_tested_interval,
                'Duration of the tested interval (in days)': (end_time_of_tested_interval - begin_time_of_tested_interval).days,
                'USD annualized ROI (from first to last trade)': dict_roi_results['portfolio_roi'],
                'Cryptocurrency annualized ROI delta (from first to last trade)': dict_roi_results['roi_delta_compared_to_benchmark'],
                'Beginning benchmark value (first to last trade)': find_price(
                    df_benchmark,
                    desired_index=(begin_time_of_tested_interval, benchmark_data_specifications['benchmark_key']),
                    boolean_allow_older_prices=False,
                    boolean_allow_newer_prices=False,
                    boolean_warnings=True,
                    boolean_errors=display_options['errors_on_benchmark_gap']
                ),
                'Ending benchmark value (first to last trade)': find_price(
                    df_benchmark,
                    desired_index=(end_time_of_tested_interval, benchmark_data_specifications['benchmark_key']),
                    boolean_allow_older_prices=False,
                    boolean_allow_newer_prices=False,
                    boolean_warnings=True,
                    boolean_errors=display_options['errors_on_benchmark_gap']
                ),
                'Initial budget': float_budget_in_usd,
                'Ending portfolio value': df_trading_journal['Portfolio value'].iloc[-1],
                'Holding period volatility': calculate_volatility(df_daily_returns, len(df_daily_returns)),
                'Annual volatility': calculate_volatility(df_daily_returns, 365),
                'Monthly volatility': calculate_volatility(df_daily_returns, 30),
                'Weekly volatility': calculate_volatility(df_daily_returns, 7),
                'Alpha': calculate_alpha(
                    annualized_portfolio_return=dict_roi_results['portfolio_roi'],
                    risk_free_rate=benchmark_data_specifications[
                        'risk_free_rate'
                    ],
                    beta_exposure=beta,
                    annualized_market_return=dict_roi_results['benchmark_roi']
                ),
                'Sharpe ratio (holding period)': calculate_sharpe_ratio(
                    df_daily_returns,
                    portfolio_roi_usd=dict_roi_results['portfolio_roi'],
                    risk_free_rate=benchmark_data_specifications[
                        'risk_free_rate'
                    ]
                ),
                'Sharpe ratio (yearly)': calculate_sharpe_ratio(
                    df_daily_returns,
                    portfolio_roi_usd=dict_roi_results['portfolio_roi'],
                    risk_free_rate=benchmark_data_specifications[
                        'risk_free_rate'
                    ],
                    days=365
                ),
                'Beta relative to benchmark': beta,
                'Maximum drawdown': maximum_drawdown,
                'Maximum drawdown duration': maximum_drawdown_duration,
                'Maximum drawdown peak date': maximum_drawdown_peak_date,
                'Maximum drawdown trough date': maximum_drawdown_trough_date,
                'Other metrics --->': 'Other metrics --->',
                'Total transaction cost': calculate_transaction_cost(df_trading_journal),
                'Number of trades': len(
                    dict_execution_results['df_trading_journal']
                ),
                'Number of unique assets traded': len(df_trading_journal['Asset'].unique()),
                'Average ticket size': average_ticket_size_absolute_value,
                'Average cash': round(
                    dict_execution_results['df_trading_journal']['Cash'].mean(),
                    2
                ),
                'Other info --->': 'Other info --->',
                'Start time': start_time,
                'End time': datetime.now(),
                'Parameter 1': strategy_hyperparameters['sell_parameter'],
                'Parameter 2': strategy_hyperparameters['buy_parameter'],
                'Comments': dict_execution_results['comments'],
                'Benchmark return metrics --->': 'Benchmark return metrics --->',
                'Benchmark USD annualized ROI (from first to last trade)': dict_roi_results['benchmark_roi'],
                'Benchmark risk metrics --->': 'Benchmark risk metrics --->',
                'Benchmark holding period volatility': 'NOT IMPLEMENTED',
                'Benchmark annual volatility': 'NOT IMPLEMENTED',
                'Benchmark monthly volatility': 'NOT IMPLEMENTED',
                'Benchmark weekly volatility': 'NOT IMPLEMENTED',
                'Benchmark Beta relative to benchmark': 'NOT IMPLEMENTED',
                'Benchmark maximum drawdown': benchmark_maximum_drawdown,
                'Benchmark maximum drawdown duration': benchmark_maximum_drawdown_duration,
                'Benchmark maximum drawdown peak date': benchmark_maximum_drawdown_peak_date,
                'Benchmark maximum drawdown trough date': benchmark_maximum_drawdown_trough_date,
                'Benchmark other metrics --->': 'NOT IMPLEMENTED',
                'Benchmark Sharpe ratio (holding period)': 'NOT IMPLEMENTED',
                'Benchmark Sharpe ratio (yearly)': 'NOT IMPLEMENTED',
            },
            ignore_index=True
        )

        plot_equity_curve(
            df_daily_returns,
            df_daily_benchmark_returns,
            display_options=display_options,
            boolean_plot=display_options['boolean_plot_equity_curve'],
            boolean_save_to_disk=display_options['boolean_save_equity_curve_to_disk'],
        )

    return df_performance

def plot_equity_curve(
        df_daily_returns,
        df_daily_benchmark_returns,
        display_options,
        boolean_plot=False,
        boolean_relative=False,
        boolean_save_to_disk=True,
    ):
    """Creates an equity based on daily returns."""
    string_directory = display_options['string_results_directory']
    result_no = len(
        [name for name in os.listdir(string_directory) if os.path.isfile(
            os.path.join(
                string_directory,
                name
            )
        )]
    ) / 2

    number_of_result_files_plus_1 = 1 + int(result_no)

    fig, ax = plt.subplots(figsize=(12, 8))
    ax.xaxis.set_major_locator(
        MonthLocator(
            bymonthday=1,
            interval=3,
            tz=None
        )
    )
    ax.xaxis.set_major_formatter(DateFormatter("%y-%m-%d"))

    plt.legend()

    df_daily_returns.plot(
        y=['portfolio_value_normalized', 'benchmark_portfolio_value_normalized'],
        label=['Portfolio value in base currency (after fees)', 'Benchmark value in base currency (no fees considered)'],
        title='Equity curve'
    )

    if boolean_save_to_disk:
        plt.savefig(string_directory + '/equity_curve_' + str(number_of_result_files_plus_1) + '.png')

    if boolean_plot:
        plt.show()



import pathlib
from setuptools import setup
HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
setup(
    name="quantbacktest",
    version="0.0.14",
    description="This backtesting is used to backtest algorithmic/quant trading strategies.",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Jan Frederic Spörer",
    author_email="jan.spoerer@whu.edu",
    license="BSD <two-clau></two-clau>se license",
    packages=["quantbacktest"],
    # install_requires=["datetime", "numpy", "matplotlib", "pandas"],
    url="https://gitlab.com/fsbc/theses/quantbacktest",
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "FullBacktest=quantbacktest.__main__:main",
        ]
    },
    include_package_data=True,
) # Sources of this setup.py: Jonathan Hsu (Medium article) and Geir Arne Hjelle (Realpython)
# https://medium.com/better-programming/how-to-publish-your-first-python-package-its-not-that-hard-6202f74f5954
# https://realpython.com/pypi-publish-python-package/



"""Provides individual metrics calculations.

This module is more low-level than _3_performance_evaluation."""

# For mathematical operations
from math import sqrt, exp, log

# For tensor calculations
from numpy import cov

# For managing tables properly
from pandas import to_datetime

from _helper_functions import find_dataframe_value_with_keywords, \
    datetime_to_string


def calculate_alpha(
        annualized_portfolio_return,
        risk_free_rate,
        beta_exposure,
        annualized_market_return
    ):
    """Calculates the Jensen's alpha of a portfolio against a benchmark."""
    market_risk_premium = annualized_market_return - risk_free_rate

    alpha = annualized_portfolio_return - (
        risk_free_rate + beta_exposure * market_risk_premium
    )

    return alpha

def calculate_beta(df_daily_returns, df_daily_benchmark_returns):
    """Calculates the beta of a portfolio against a benchmark."""
    portfolio_returns = df_daily_returns['relative_portfolio_return'].to_numpy()
    benchmark_returns = df_daily_benchmark_returns['benchmark_relative_return'].to_numpy()

    covariance_matrix = cov(portfolio_returns, benchmark_returns)

    benchmark_portfolio_covariance = covariance_matrix[0][1]
    benchmark_variance = covariance_matrix[1][1]

    return benchmark_portfolio_covariance / benchmark_variance

def calculate_maximum_drawdown(
        df_daily_returns,
        column_with_portfolio_values
    ):
    """Calculates the maximum_drawdown of a portfolio using portfolio_value from df_daily_returns.

    Outputs the maximum_drawdown ratio.
    """
    dict_peak_tracking = {
        'first_peak': {
            'peak': {
                column_with_portfolio_values: df_daily_returns[column_with_portfolio_values].iloc[0],
                'datetime': df_daily_returns.index.values[0]
            },
            'trough': {
                column_with_portfolio_values: df_daily_returns[column_with_portfolio_values].iloc[0],
                'datetime': df_daily_returns.index.values[0]
            }
        },
        'second_peak': {
            'peak': {
                column_with_portfolio_values: df_daily_returns[column_with_portfolio_values].iloc[0],
                'datetime': df_daily_returns.index.values[0]
            },
            'trough': {
                column_with_portfolio_values: df_daily_returns[column_with_portfolio_values].iloc[0],
                'datetime': df_daily_returns.index.values[0]
            }
        }
    }

    for datetime, row in df_daily_returns.iterrows():
        if row[column_with_portfolio_values] > dict_peak_tracking['second_peak']['peak'][column_with_portfolio_values]:
            dict_peak_tracking['second_peak']['peak'][column_with_portfolio_values] = row[column_with_portfolio_values]
            dict_peak_tracking['second_peak']['peak']['datetime'] = datetime
            dict_peak_tracking['second_peak']['trough'][column_with_portfolio_values] = row[column_with_portfolio_values]
            dict_peak_tracking['second_peak']['trough']['datetime'] = datetime
        if row[column_with_portfolio_values] < dict_peak_tracking['second_peak']['trough'][column_with_portfolio_values]:
            dict_peak_tracking['second_peak']['trough'][column_with_portfolio_values] = row[column_with_portfolio_values]
            dict_peak_tracking['second_peak']['trough']['datetime'] = datetime
        if row[column_with_portfolio_values] < dict_peak_tracking['first_peak']['trough'][column_with_portfolio_values]:
            dict_peak_tracking['first_peak']['trough'][column_with_portfolio_values] = row[column_with_portfolio_values]
            dict_peak_tracking['first_peak']['trough']['datetime'] = datetime

        drawdown_second = (
            dict_peak_tracking['second_peak']['peak'][column_with_portfolio_values] - dict_peak_tracking['second_peak']['trough'][column_with_portfolio_values]
        ) / dict_peak_tracking['second_peak']['peak'][column_with_portfolio_values]
        drawdown_first = (
            dict_peak_tracking['first_peak']['peak'][column_with_portfolio_values] - dict_peak_tracking['first_peak']['trough'][column_with_portfolio_values]
        ) / dict_peak_tracking['first_peak']['peak'][column_with_portfolio_values]
        if drawdown_second > drawdown_first:
            dict_peak_tracking['first_peak'] = dict_peak_tracking['second_peak']

    maximum_drawdown_duration = dict_peak_tracking['first_peak']['trough']['datetime'] - dict_peak_tracking['first_peak']['peak']['datetime']
    peak_date = dict_peak_tracking['first_peak']['peak']['datetime']
    trough_date = dict_peak_tracking['first_peak']['trough']['datetime']

    return drawdown_first, maximum_drawdown_duration, peak_date, trough_date

def calculate_roi(
        df_trading_journal,
        float_budget_in_usd,
        df_benchmark=None,
        df_price_column_name='price',
        df_time_column_name='datetime',
        df_benchmark_price_column_name='price',
        df_benchmark_time_column_name='datetime',
        df_trading_journal_price_column_name='Portfolio value',
        df_trading_journal_time_column_name='datetime'
    ):
    """Calculates annualized returns (on equity, not total assets).

    It can calculate standalone returns without benchmarking (in USD/fiat) or
    returns in relation to a benchmark. It can also use different start and end
    points for calculating the time frame, i.e., the duration between the first
    and the last trade OR the duration between the first and the last data point
    of the price data.
    """
    start_time = df_trading_journal[
        df_trading_journal_time_column_name
    ].iloc[0]
    end_time = df_trading_journal[
        df_trading_journal_time_column_name
    ].iloc[-1]

    portfolio_roi = calculate_roi_math(
        end_value=df_trading_journal[
            df_trading_journal_price_column_name
    ].iloc[-1],
        start_value=float_budget_in_usd,
        end_time=end_time,
        start_time=start_time
    )

    roi_delta_compared_to_benchmark = None
    benchmark_roi = None

    if df_benchmark is not None:
        end_value_benchmark = find_dataframe_value_with_keywords(
            df_benchmark,
            search_term_1=end_time,
            search_column_name_1=df_benchmark_time_column_name,
            search_term_2=None,
            search_column_name_2=None,
            output_column_name=df_benchmark_price_column_name,
            first_last_or_all_elements='First'
        )

        begin_value_benchmark = find_dataframe_value_with_keywords(
            df_benchmark,
            search_term_1=start_time,
            search_column_name_1=df_benchmark_time_column_name,
            search_term_2=None,
            search_column_name_2=None,
            output_column_name=df_benchmark_price_column_name,
            first_last_or_all_elements='First'
        )

        benchmark_roi = calculate_roi_math(
            end_value=end_value_benchmark,
            start_value=begin_value_benchmark,
            end_time=end_time,
            start_time=start_time
        )

        roi_delta_compared_to_benchmark = portfolio_roi - benchmark_roi

    return {
        'portfolio_roi': portfolio_roi,
        'benchmark_roi': benchmark_roi,
        'roi_delta_compared_to_benchmark': roi_delta_compared_to_benchmark
    }

def calculate_roi_math(
        end_value,
        start_value,
        end_time,
        start_time
    ):
    unadjusted_return_factor = end_value / start_value
    time_duration = (end_time - start_time).total_seconds() / 86400

    roi = exp(
        log(
            unadjusted_return_factor
        ) * 365 / (time_duration)
    ) - 1

    return roi

def calculate_sharpe_ratio(
        df_daily_returns,
        portfolio_roi_usd,
        days=None,
        risk_free_rate=None
    ):
    """Calculates the Sharpe ratio of a given set of returns."""
    volatility = calculate_volatility(df_daily_returns, days)

    if risk_free_rate is None:
        sharpe_ratio = portfolio_roi_usd / volatility
    else:
        sharpe_ratio = (portfolio_roi_usd - risk_free_rate) / volatility
    return sharpe_ratio


def calculate_transaction_cost(df_trading_journal):
    return round(
        sum(df_trading_journal["Total fees (as absolute)"]),
        2
    )

def calculate_volatility(df_daily_returns, time_adjustment_in_days=None):
    """Calculates the volatility of a portfolio using daily portfolio returns.

    Outputs the volatility over the given time series by default; can also be
    adjusted for any arbitrary time using the time_adjustment_in_days parameter.
    """
    if time_adjustment_in_days is None:
        time_adjustment_in_days = len(df_daily_returns)

    volatility = df_daily_returns['relative_portfolio_return'].std() * sqrt(
        time_adjustment_in_days
    ) / sqrt(len(df_daily_returns))

    if round(volatility, 4) == 0:
        raise ValueError('Volatility cannot be zero or close to zero. Please check if daily returns are correctly calculated and if any trades were made. \n{df_daily_returns}')

    return volatility



"""Serves as a high-level caller."""

# For counting the number of result files in the "results" folder to allow for
# consistent file versioning when saving the results as a csv file.
import os
import os.path

# For managing dates
from datetime import datetime

# For vector operations
from numpy import array

# For plots (especially the heatmap)
import matplotlib.pyplot as plt

from _1_data_preparation import prepare_data, save_dataframe_to_csv
from _2_strategy_execution import test_strategies


def backtest(
        file_path_with_price_data,
        int_chosen_strategy,
        dict_crypto_options,
        float_budget_in_usd,
        margin_loan_rate,
        display_options,
        general_settings,
        constraints,
        start_time,
        comments,
        file_path_with_signal_data=None,
        file_path_with_token_data=None,# Only for multi-asset strategies.
        name_of_foreign_key_in_price_data_table=None,
        name_of_foreign_key_in_token_metadata_table=None,
        boolean_allow_shorting=False,
        list_trading_execution_delay_after_signal_in_hours={
            'delay_before_buying': 0,
            'delay_before_selling': 0
        },
        minimum_expected_mispricing_trigger_in_percent={
            'mispricing_when_buying': 0.0,
            'mispricing_when_selling': 0.0
        },
        strategy_hyperparameters=None,
        sell_at_the_end=True,
        list_times_of_split_for_robustness_test=None,
        benchmark_data_specifications={# Bitcoin as default benchmark
            'name_of_column_with_benchmark_primary_key': 'id',
            'benchmark_key': 'TP3B-248N-Q',
            'file_path_with_benchmark_data': 'raw_itsa_data/20190717_itsa_tokenbase_top600_wtd302_token_daily.csv'
        }
    ):
    """Backtests a user-defined strategy.

    With user-defined price data and possibly many other parameters.
    """
    # assert percentage_selling_fees_and_spread >= 0, "Please choose positive selling slippage."
    # assert percentage_buying_fees_and_spread >= 0, "Please choose positive buying slippage."
    # assert percentage_selling_fees_and_spread == 0, "Please choose a selling slippage greater than zero."
    # assert percentage_buying_fees_and_spread == 0, "Please choose a buying slippage greater than zero."
    assert float_budget_in_usd > 0, "Please choose positive budget."
    # assert list_trading_execution_delay_after_signal_in_hours >= 0, "Please choose an execution delay greater than zero."
    # assert list_trading_execution_delay_after_signal_in_hours == 0, "Please choose an exectution delay greater than zero."

    ## Check if exposure is > 0 and <= 100
    ## Check if margin is between 0.01 and 1
    ## Check if fees are >= 0 and warn if fees = 0

    df_prices = prepare_data(
        file_path_with_price_data=file_path_with_price_data,
        file_path_with_token_data=file_path_with_token_data,
        strategy_hyperparameters=strategy_hyperparameters,
        name_of_foreign_key_in_price_data_table=name_of_foreign_key_in_price_data_table,
        name_of_foreign_key_in_token_metadata_table=name_of_foreign_key_in_token_metadata_table,
    )

    if list_times_of_split_for_robustness_test is None:
        dfs_results = test_strategies(
            df_prices=df_prices,
            int_chosen_strategy=int_chosen_strategy,
            float_budget_in_usd=float_budget_in_usd,
            file_path_with_signal_data=file_path_with_signal_data,
            margin_loan_rate=margin_loan_rate,
            boolean_allow_shorting=boolean_allow_shorting,
            list_trading_execution_delay_after_signal_in_hours=list_trading_execution_delay_after_signal_in_hours,
            dict_crypto_options=dict_crypto_options,
            minimum_expected_mispricing_trigger_in_percent=minimum_expected_mispricing_trigger_in_percent,
            strategy_hyperparameters=strategy_hyperparameters,
            sell_at_the_end=sell_at_the_end,
            benchmark_data_specifications=benchmark_data_specifications,
            display_options=display_options,
            general_settings=general_settings,
            start_time=start_time,
            comments=comments
        )

    else:
        for sell_parameter in strategy_hyperparameters['sell_parameter_space']:
            for buy_parameter in strategy_hyperparameters['buy_parameter_space']:
                strategy_hyperparameters['sell_parameter'] = sell_parameter/10
                strategy_hyperparameters['buy_parameter'] = buy_parameter/10

                for index, dates in enumerate(list_times_of_split_for_robustness_test):
                    strategy_hyperparameters['start_time'] = dates[0]
                    strategy_hyperparameters['end_time'] = dates[1]

                    dfs_intermediate_results = test_strategies(
                        df_prices,
                        int_chosen_strategy=int_chosen_strategy,
                        float_budget_in_usd=float_budget_in_usd,
                        file_path_with_signal_data=file_path_with_signal_data,
                        margin_loan_rate=margin_loan_rate,
                        boolean_allow_shorting=boolean_allow_shorting,
                        list_trading_execution_delay_after_signal_in_hours=list_trading_execution_delay_after_signal_in_hours,
                        dict_crypto_options=dict_crypto_options,
                        minimum_expected_mispricing_trigger_in_percent=minimum_expected_mispricing_trigger_in_percent,
                        strategy_hyperparameters=strategy_hyperparameters,
                        sell_at_the_end=sell_at_the_end,
                        benchmark_data_specifications=benchmark_data_specifications,
                        display_options=display_options,
                        constraints=constraints,
                        general_settings=general_settings,
                        start_time=start_time,
                        comments=comments
                    )

                    try:
                        dfs_results[0] = dfs_results[0].append(
                            dfs_intermediate_results[0]
                        )
                        dfs_results[1] = dfs_results[1].append(
                            dfs_intermediate_results[1]
                        )
                    except:
                        dfs_results = dfs_intermediate_results

    save_dataframe_to_csv(
        dfs_results[0],
        string_name='backtesting_result_metrics',
        string_directory=display_options['string_results_directory'],
    )

    comments = {
        **{
            'file_path_with_price_data': file_path_with_price_data,
            'file_path_with_token_data': file_path_with_token_data,
            'name_of_foreign_key_in_price_dat_table': name_of_foreign_key_in_price_data_table,
            'name_of_foreign_key_in_token_metadata_table': name_of_foreign_key_in_token_metadata_table,
            'float_budget_in_usd': float_budget_in_usd,
            'margin_loan_rate': margin_loan_rate,
            'boolean_allow_shorting': boolean_allow_shorting,
            'list_trading_execution_delay_after_signal_in_hours': list_trading_execution_delay_after_signal_in_hours,
            'dict_crypto_options': dict_crypto_options,
            'minimum_expected_mispricing_trigger_in_percent': minimum_expected_mispricing_trigger_in_percent,
            'strategy_hyperparameters': strategy_hyperparameters,
            'sell_at_the_end': sell_at_the_end,
        },
        **comments
    }


    return {
        'df_performance': dfs_results[0],
        'df_trading_journal': dfs_results[1],
        'comments': comments
    }

def plot_robustness_heatmap(
        df_performance,
        display_options,
        boolean_save=True,
        boolean_show=False
    ):
    """Plots each row that is contained in the performance table.

    With the attributes 'start date' on the x-axis and 'duration' on the y-axis.
    """

    dict_heatmap = {
        'Begin time of tested interval': [],
        'Duration of the tested interval': [],
        'USD annualized ROI (from first to last trade)': []
    }

    for index, row in df_performance.iterrows():
        begin_date = row['Begin time of tested interval']
        strategy_duration_in_days = int(
            row['Duration of the tested interval'].days
        )
        roi = row['USD annualized ROI (from first to last trade)']
        dict_heatmap['Begin time of tested interval'].append(begin_date)
        dict_heatmap['Duration of the tested interval'].append(strategy_duration_in_days)
        dict_heatmap['USD annualized ROI (from first to last trade)'].append(roi)

    # x_axis = array(dict_heatmap['Begin time of tested interval'], dtype=datetime)
    x_axis = [i.to_pydatetime() for i in dict_heatmap['Begin time of tested interval']]
    y_axis = dict_heatmap['Duration of the tested interval']
    z_values = dict_heatmap['USD annualized ROI (from first to last trade)']

    figure, axis = plt.subplots()
    plt.ylabel('Duration (in days)')
    axis.set_xlabel('Begin time of tested interval')
    plt.xticks(rotation=70)
    plt.title('ROI heatmap (in percent)')
    points = axis.scatter(
        x_axis,
        y_axis,
        c=z_values,
        s=1300,# Size of scatters
        cmap='RdYlGn',# Taken from here: https://matplotlib.org/users/colormaps.html
        marker="s"
    )
    figure.colorbar(points)

    string_directory = display_options['string_results_directory']
    result_no = len(
        [name for name in os.listdir(string_directory) if os.path.isfile(
            os.path.join(
                string_directory,
                name
            )
        )]
    ) / 2

    number_of_result_files_plus_1 = 1 + int(result_no)

    if boolean_save:
        plt.savefig(string_directory + '/robustness_heatmap_' + str(number_of_result_files_plus_1) + '.png')

    if boolean_show:
        plt.show()

def backtest_visualizer(
        file_path_with_price_data,
        int_chosen_strategy,
        dict_crypto_options,
        benchmark_data_specifications,
        display_options,
        strategy_hyperparameters,
        constraints,
        general_settings,
        margin_loan_rate,
        file_path_with_token_data=None, # Only for multi-asset strategies.
        name_of_foreign_key_in_price_data_table=None,
        name_of_foreign_key_in_token_metadata_table=None,
        float_budget_in_usd=10000,
        boolean_allow_shorting=False,
        list_trading_execution_delay_after_signal_in_hours={
            'delay_before_buying': 24,
            'delay_before_selling': 24
        },
        minimum_expected_mispricing_trigger_in_percent={
            'mispricing_when_buying': 0.0,
            'mispricing_when_selling': 0.0
        },
        sell_at_the_end=True,
        list_times_of_split_for_robustness_test=None,
        file_path_with_signal_data=None,
        comments={}
    ):
    """Prints and plots results from the performance table."""

    start_time = datetime.now()
    dict_backtesting = backtest(
        file_path_with_price_data=file_path_with_price_data,
        file_path_with_token_data=file_path_with_token_data,
        name_of_foreign_key_in_price_data_table=name_of_foreign_key_in_price_data_table,
        name_of_foreign_key_in_token_metadata_table=name_of_foreign_key_in_token_metadata_table,
        float_budget_in_usd=float_budget_in_usd,
        margin_loan_rate=margin_loan_rate,
        boolean_allow_shorting=boolean_allow_shorting,
        list_trading_execution_delay_after_signal_in_hours=list_trading_execution_delay_after_signal_in_hours,
        dict_crypto_options=dict_crypto_options,
        minimum_expected_mispricing_trigger_in_percent=minimum_expected_mispricing_trigger_in_percent,
        strategy_hyperparameters=strategy_hyperparameters,
        sell_at_the_end=sell_at_the_end,
        list_times_of_split_for_robustness_test=list_times_of_split_for_robustness_test,
        benchmark_data_specifications=benchmark_data_specifications,
        int_chosen_strategy=int_chosen_strategy,
        file_path_with_signal_data=file_path_with_signal_data,
        display_options=display_options,
        constraints=constraints,
        general_settings=general_settings,
        start_time=start_time,
        comments=comments
    )
    end_time = datetime.now()
    elapsed_time = end_time - start_time
    asset_name = list(dict_crypto_options.keys())[0]
    print("\n")
    print("Execution started at: " + str(start_time) + ", finished at: " + str(end_time) + ", elapsed time:", str(elapsed_time.total_seconds()) + "s")
    print("****  Performance overview ->", asset_name, "<- ****")
    print("Key metrics")
    print("*USD annualized ROI (from first to last trade):", "{0:.2%}".format(dict_backtesting['df_performance']['USD annualized ROI (from first to last trade)'].iloc[-1]))
    if type(dict_backtesting['df_performance']['Cryptocurrency annualized ROI delta (from first to last trade)'].iloc[-1]) is str:
        print("*Cryptocurrency annualized ROI delta (from first to last trade):", dict_backtesting['df_performance']['Cryptocurrency annualized ROI delta (from first to last trade)'].iloc[-1])
    else:
        print("*Cryptocurrency annualized ROI delta (from first to last trade):", "{0:.2%}".format(dict_backtesting['df_performance']['Cryptocurrency annualized ROI delta (from first to last trade)'].iloc[-1]))
    print("Number of trades:", len(dict_backtesting['df_trading_journal']))
    print("Other metrics")
    print("**** Assumptions ****")
    print("*Budget:", dict_backtesting['comments']['float_budget_in_usd'])
    print("*Check arguments and parameter defaults for a full list of assumptions.*")

    plot_robustness_heatmap(
        dict_backtesting['df_performance'],
        display_options=display_options,
        boolean_show=display_options['boolean_plot_heatmap']
    )

    return dict_backtesting



__version__ = "0.0.14"

# Importing modules from this repository
import sys
PATH_TO_PROJECT = '/home/janspoerer/code/janspoerer/quantbacktest'
sys.path.insert(1, PATH_TO_PROJECT + '/quantbacktest/components')

from _0_wrappers import backtest_visualizer
